package ru.dbotthepony.kstarbound.client.render.entity

import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import ru.dbotthepony.kstarbound.client.world.ClientChunk
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.world.entities.Entity
import ru.dbotthepony.kvector.arrays.Matrix4fStack
import ru.dbotthepony.kvector.vector.Vector2d

/**
 * Базовый класс, отвечающий за отрисовку определённого ентити в мире
 *
 * Считается, что процесс отрисовки ограничен лишь одним слоем (т.е. отрисовка происходит в один проход)
 */
open class EntityRenderer(val client: StarboundClient, val entity: Entity, open var chunk: ClientChunk?) {
	open val renderPos: Vector2d get() = entity.position

	open fun render(stack: Matrix4fStack) {

	}

	open val layer: Int get() = Z_LEVEL_ENTITIES

	companion object {
		/**
		 * Pseudo Z position for entities, for them to appear behind tile geometry,
		 * but in front of background walls geometry
		 */
		const val Z_LEVEL_ENTITIES = 30000

		private val renderers = Reference2ObjectOpenHashMap<Class<*>, (client: StarboundClient, entity: Entity, chunk: ClientChunk?) -> EntityRenderer>()

		@Suppress("unchecked_cast")
		fun <T : Entity> registerRenderer(clazz: Class<T>, renderer: (client: StarboundClient, entity: T, chunk: ClientChunk?) -> EntityRenderer) {
			check(renderers.put(clazz, renderer as (client: StarboundClient, entity: Entity, chunk: ClientChunk?) -> EntityRenderer) == null) { "Already has renderer for ${clazz.canonicalName}!" }
		}

		inline fun <reified T : Entity> registerRenderer(noinline renderer: (client: StarboundClient, entity: T, chunk: ClientChunk?) -> EntityRenderer) {
			registerRenderer(T::class.java, renderer)
		}

		fun getRender(client: StarboundClient, entity: Entity, chunk: ClientChunk? = null): EntityRenderer {
			val factory = renderers[entity::class.java] ?: return EntityRenderer(client, entity, chunk)
			return factory.invoke(client, entity, chunk)
		}

		init {
			registerRenderer(::ItemRenderer)
		}
	}
}

