package ru.dbotthepony.kstarbound.defs

import com.google.common.collect.ImmutableSet
import ru.dbotthepony.kstarbound.io.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.io.json.builder.JsonImplementation
import ru.dbotthepony.kstarbound.util.KOptional

@JsonImplementation(BaseMovementParameters.Impl::class)
interface BaseMovementParameters {
	val mass: KOptional<Float>
	val gravityMultiplier: KOptional<Float>
	val liquidBuoyancy: KOptional<Float>
	val airBuoyancy: KOptional<Float>
	val bounceFactor: KOptional<Float>

	// If set to true, during an update that has more than one internal movement
	// step, the movement will stop on the first bounce.
	val stopOnFirstBounce: KOptional<Boolean>

	// Cheat when sliding on the ground, by trying to correct upwards before
	// other directions (within a set limit).  Allows smooth sliding along
	// horizontal ground without losing horizontal speed.
	val enableSurfaceSlopeCorrection: KOptional<Boolean>
	val slopeSlidingFactor: KOptional<Float>
	val maxMovementPerStep: KOptional<Float>
	val maximumCorrection: KOptional<Float>
	val speedLimit: KOptional<Float>

	val stickyCollision: KOptional<Boolean>
	val stickyForce: KOptional<Float>

	val airFriction: KOptional<Float>
	val liquidFriction: KOptional<Float>
	val groundFriction: KOptional<Float>

	val collisionEnabled: KOptional<Boolean>
	val frictionEnabled: KOptional<Boolean>
	val gravityEnabled: KOptional<Boolean>

	val maximumPlatformCorrection: KOptional<Float>
	val maximumPlatformCorrectionVelocityFactor: KOptional<Float>

	val physicsEffectCategories: KOptional<ImmutableSet<String>>

	@JsonFactory
	data class Impl(
		override val mass: KOptional<Float> = KOptional.empty(),
		override val gravityMultiplier: KOptional<Float> = KOptional.empty(),
		override val liquidBuoyancy: KOptional<Float> = KOptional.empty(),
		override val airBuoyancy: KOptional<Float> = KOptional.empty(),
		override val bounceFactor: KOptional<Float> = KOptional.empty(),
		override val stopOnFirstBounce: KOptional<Boolean> = KOptional.empty(),
		override val enableSurfaceSlopeCorrection: KOptional<Boolean> = KOptional.empty(),
		override val slopeSlidingFactor: KOptional<Float> = KOptional.empty(),
		override val maxMovementPerStep: KOptional<Float> = KOptional.empty(),
		override val maximumCorrection: KOptional<Float> = KOptional.empty(),
		override val speedLimit: KOptional<Float> = KOptional.empty(),
		override val stickyCollision: KOptional<Boolean> = KOptional.empty(),
		override val stickyForce: KOptional<Float> = KOptional.empty(),
		override val airFriction: KOptional<Float> = KOptional.empty(),
		override val liquidFriction: KOptional<Float> = KOptional.empty(),
		override val groundFriction: KOptional<Float> = KOptional.empty(),
		override val collisionEnabled: KOptional<Boolean> = KOptional.empty(),
		override val frictionEnabled: KOptional<Boolean> = KOptional.empty(),
		override val gravityEnabled: KOptional<Boolean> = KOptional.empty(),
		override val maximumPlatformCorrection: KOptional<Float> = KOptional.empty(),
		override val maximumPlatformCorrectionVelocityFactor: KOptional<Float> = KOptional.empty(),
		override val physicsEffectCategories: KOptional<ImmutableSet<String>> = KOptional.empty(),
	) : BaseMovementParameters {
		fun merge(other: Impl): Impl {
			return Impl(
				mass = mass.or(other.mass),
				gravityMultiplier = gravityMultiplier.or(other.gravityMultiplier),
				liquidBuoyancy = liquidBuoyancy.or(other.liquidBuoyancy),
				airBuoyancy = airBuoyancy.or(other.airBuoyancy),
				bounceFactor = bounceFactor.or(other.bounceFactor),
				stopOnFirstBounce = stopOnFirstBounce.or(other.stopOnFirstBounce),
				enableSurfaceSlopeCorrection = enableSurfaceSlopeCorrection.or(other.enableSurfaceSlopeCorrection),
				slopeSlidingFactor = slopeSlidingFactor.or(other.slopeSlidingFactor),
				maxMovementPerStep = maxMovementPerStep.or(other.maxMovementPerStep),
				maximumCorrection = maximumCorrection.or(other.maximumCorrection),
				speedLimit = speedLimit.or(other.speedLimit),
				stickyCollision = stickyCollision.or(other.stickyCollision),
				stickyForce = stickyForce.or(other.stickyForce),
				airFriction = airFriction.or(other.airFriction),
				liquidFriction = liquidFriction.or(other.liquidFriction),
				groundFriction = groundFriction.or(other.groundFriction),
				collisionEnabled = collisionEnabled.or(other.collisionEnabled),
				frictionEnabled = frictionEnabled.or(other.frictionEnabled),
				gravityEnabled = gravityEnabled.or(other.gravityEnabled),
				maximumPlatformCorrection = maximumPlatformCorrection.or(other.maximumPlatformCorrection),
				maximumPlatformCorrectionVelocityFactor = maximumPlatformCorrectionVelocityFactor.or(other.maximumPlatformCorrectionVelocityFactor),
				physicsEffectCategories = physicsEffectCategories.or(other.physicsEffectCategories),
			)
		}
	}
}