package ru.dbotthepony.kstarbound.defs.particle

import com.google.common.collect.ImmutableList
import ru.dbotthepony.kstarbound.io.json.builder.JsonImplementation
import ru.dbotthepony.kstarbound.util.VirtualProperty
import ru.dbotthepony.kvector.vector.RGBAColor
import ru.dbotthepony.kvector.vector.Vector2d
import ru.dbotthepony.kvector.vector.Vector4d

@JsonImplementation(ParticleVariance::class)
interface IParticleVariance {
	val offset: Vector2d?
	val position: Vector2d?
	val offsetRegion: Vector4d?
	val initialVelocity: Vector2d?
	val approach: Vector2d?
	val angularVelocity: Double?
	val size: Double?
	val color: RGBAColor?

	companion object {
		fun chain(vararg particles: IParticleVariance): IParticleVariance {
			@Suppress("name_shadowing")
			val particles = ImmutableList.copyOf(particles)

			return object : IParticleVariance {
				override val offset by VirtualProperty(IParticleVariance::offset, particles)
				override val position by VirtualProperty(IParticleVariance::position, particles)
				override val offsetRegion by VirtualProperty(IParticleVariance::offsetRegion, particles)
				override val initialVelocity by VirtualProperty(IParticleVariance::initialVelocity, particles)
				override val approach by VirtualProperty(IParticleVariance::approach, particles)
				override val angularVelocity by VirtualProperty(IParticleVariance::angularVelocity, particles)
				override val size by VirtualProperty(IParticleVariance::size, particles)
				override val color by VirtualProperty(IParticleVariance::color, particles)
			}
		}
	}
}
