package ru.dbotthepony.kstarbound.io.json

import com.google.gson.Gson
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.util.KOptional
import java.lang.reflect.ParameterizedType

@Suppress("DEPRECATION")
object KOptionalTypeAdapter : TypeAdapterFactory {
	override fun <T : Any> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
		if (type.rawType === KOptional.Nullable::class.java || type.rawType === KOptional::class.java || type.rawType === KOptional.NotNull::class.java) {
			val notnull = type.rawType === KOptional.NotNull::class.java
			val param = (type.type as? ParameterizedType ?: return null).actualTypeArguments[0]
			val token = TypeToken.get(param)
			val isBool = token.rawType === Boolean::class.java

			return object : TypeAdapter<KOptional<Any?>>() {
				private val adapter0 = gson.getAdapter(token) as TypeAdapter<Any?>

				override fun write(out: JsonWriter, value: KOptional<Any?>?) {
					if (value === null) {
						out.nullValue()
					} else if (value.isPresent) {
						adapter0.write(out, value.value)
					}
				}

				override fun read(`in`: JsonReader): KOptional<Any?> {
					if (isBool) {
						if (notnull) {
							return KOptional((adapter0.read(`in`) ?: throw JsonSyntaxException("This KOptional does not accept nulls")) as Boolean) as KOptional<Any?>
						} else {
							return KOptional(adapter0.read(`in`) as Boolean?) as KOptional<Any?>
						}
					} else {
						if (notnull) {
							return KOptional(adapter0.read(`in`) ?: throw JsonSyntaxException("This KOptional does not accept nulls"))
						} else {
							return KOptional(adapter0.read(`in`))
						}
					}
				}
			} as TypeAdapter<T>
		} else {
			return null
		}
	}
}
