package ru.dbotthepony.kstarbound.world.api

import ru.dbotthepony.kvector.api.IStruct2i
import ru.dbotthepony.kvector.vector.Vector2i

interface ICellAccess {
	/**
	 * non-null - valid cell (maybe wrapped around)
	 * null - invalid cell (outside world bounds)
	 */
	fun getCell(x: Int, y: Int): IChunkCell?
	fun getCell(pos: IStruct2i) = getCell(pos.component1(), pos.component2())

	/**
	 * non-null - valid cell and not wrapped around
	 * null - invalid cell (outside world bounds)
	 */
	fun getCellDirect(x: Int, y: Int): IChunkCell?
	fun getCellDirect(pos: IStruct2i) = getCellDirect(pos.component1(), pos.component2())

	/**
	 * Возвращает псевдослучайное Long для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomLongFor(x: Int, y: Int): Long {
		var long = x * 738548L + y * 2191293543L
		long = long xor 8339437585692L
		long = (long ushr 4) or (long shl 52)
		long *= 7848344324L
		long = (long ushr 12) or (long shl 44)
		return long
	}

	/**
	 * Возвращает псевдослучайное нормализированное Double для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomDoubleFor(x: Int, y: Int): Double {
		return (randomLongFor(x, y) / 9.223372036854776E18) / 2.0 + 0.5
	}

	fun randomLongFor(pos: Vector2i) = randomLongFor(pos.x, pos.y)
	fun randomDoubleFor(pos: Vector2i) = randomDoubleFor(pos.x, pos.y)
}

class OffsetCellAccess(private val parent: ICellAccess, var x: Int, var y: Int) : ICellAccess {
	constructor(parent: ICellAccess, offset: IStruct2i) : this(parent, offset.component1(), offset.component2())

	override fun getCell(x: Int, y: Int): IChunkCell? {
		return parent.getCell(x + this.x, y + this.y)
	}

	override fun getCellDirect(x: Int, y: Int): IChunkCell? {
		return parent.getCellDirect(x + this.x, y + this.y)
	}

	override fun randomLongFor(x: Int, y: Int) = parent.randomLongFor(x + this.x, y + this.y)
	override fun randomDoubleFor(x: Int, y: Int) = parent.randomDoubleFor(x + this.x, y + this.y)
	override fun randomLongFor(pos: Vector2i) = parent.randomLongFor(pos.x + this.x, pos.y + this.y)
	override fun randomDoubleFor(pos: Vector2i) = parent.randomDoubleFor(pos.x + this.x, pos.y + this.y)
}
